# -*- coding: utf-8 -*-
import logging
import os
from _socket import gethostname
from importlib import import_module

from celery import Celery
from celery.bin.beat import beat as beat_bin
from celery.bin.worker import worker as worker_bin
from clinner.command import command
from clinner.exceptions import ImproperlyConfigured
from clinner.run import Main
from flower.command import FlowerCommand

from task_dispatcher.celery import app
from task_dispatcher.register import register
from task_dispatcher.settings import settings

logger = logging.getLogger(__name__)

_worker_bin = worker_bin(app=Celery())
_beat_bin = beat_bin(app=Celery())

SHOW_JSON = 'json'
SHOW_YAML = 'yaml'
SHOW_CHOICES = (SHOW_JSON, SHOW_YAML)


@command(args=_worker_bin.add_arguments, parser_opts={'help': 'Run a consumer.'})
def consumer(*args, **kwargs):
    """
    Run a consumer process.
    """
    kwargs['queues'] = kwargs.get('queues') or ['consumer']
    kwargs['hostname'] = kwargs.get('hostname') or 'consumer@{}'.format(gethostname())
    worker = app.Worker(**kwargs)
    worker.start()
    return worker.exitcode


@command(args=_worker_bin.add_arguments, parser_opts={'help': 'Run a producer.'})
def producer(*args, **kwargs):
    """
    Run a producer process.
    """
    kwargs['queues'] = kwargs.get('queues') or ['producer']
    kwargs['hostname'] = kwargs.get('hostname') or 'producer@{}'.format(gethostname())
    worker = app.Worker(**kwargs)
    worker.start()
    return worker.exitcode


@command(args=_beat_bin.add_arguments, parser_opts={'help': 'Run the scheduler.'})
def scheduler(*args, **kwargs):
    """
    Run a scheduler process.
    """
    beat = app.Beat(**kwargs)

    # Remove old startup tasks
    startup_tasks = [t[0] for t in settings.run_at_startup]
    inspect = app.control.inspect()
    scheduled = {(t['id'], t['name']) for tasks in inspect.scheduled().values() for t in tasks}
    active = {(t['id'], t['name']) for tasks in inspect.active().values() for t in tasks}
    reserved = {(t['id'], t['name']) for tasks in inspect.reserved().values() for t in tasks}
    app.control.revoke([i for i, name in (scheduled | active | reserved) if name in startup_tasks])

    # Load startup tasks
    for task_path, task_args, task_kwargs in settings.run_at_startup:
        try:
            task_module, task_name = task_path.rsplit('.', 1)
            task = getattr(import_module(task_module), task_name)
        except:
            logger.error('Cannot load task "%s"', task_path)
        else:
            task.delay(*task_args, **task_kwargs)

    return beat.run()


@command(args=((('-f', '--format'), {'choices': SHOW_CHOICES, 'default': SHOW_YAML}),),
         parser_opts={'help': 'Lists all producers and consumers registered.'})
def show(*args, **kwargs):
    """
    Lists all producers and consumers registered.
    """
    if kwargs.get('format') == SHOW_YAML:
        print(register.to_yaml())
    else:
        print(register.to_json())


@command(parser_opts={'help': 'Run Flower monitoring tool.'})
def flower(*args, **kwargs):
    """
    Run Flower monitoring tool.
    """
    args = ('flower',) + args
    flower_cmd = FlowerCommand(app=app)
    flower_cmd.execute_from_commandline(argv=args)


class TaskDispatcherCommand(Main):
    """
    Command for running producer, consumer and scheduler processes along with some other utilities. This acts as a
    Command Line Interface for Task dispatcher.
    """
    description = 'Task dispatcher command that provides a common entry point for running the different processes as ' \
                  'well as some utilities.'

    def inject_app_settings(self):
        if self.settings:
            os.environ['TASK_DISPATCHER_SETTINGS'] = self.settings

        if 'TASK_DISPATCHER_SETTINGS' not in os.environ:
            raise ImproperlyConfigured('Settings not defined')



# -*- coding: utf-8 -*-
"""
Settings.
"""
import os
from functools import partial, update_wrapper
from importlib import import_module

__all__ = ['settings']


class reset:  # noqa
    def __init__(self, func, *args, **kwargs):
        self.func = func
        update_wrapper(self, func)

    def __get__(self, instance, owner=None):
        self.instance = instance
        return partial(self, instance)

    def __call__(self, *args, **kwargs):
        self.instance.reset_default()
        self.func(*args, **kwargs)


class Settings:
    run_at_startup = []

    def __init__(self):
        self.reset_default()
        module_path = os.environ.get('TASK_DISPATCHER_SETTINGS')
        if module_path:
            self.build_from_module(module_path)

    def reset_default(self):
        """
        Reset settings to default values.
        """
        self.run_at_startup = []

    @staticmethod
    def import_settings(path):
        """
        Import a settings module that can be a module or an object.
        To import a module, his full path should be specified: *package.settings*.
        To import an object, his full path and object name should be specified: *project.settings:SettingsObject*.

        :param path: Settings full path.
        :return: Settings module or object.
        """
        try:
            try:
                m, c = path.rsplit(':', 1)
                module = import_module(m)
                s = getattr(module, c)
            except ValueError:
                s = import_module(path)
        except ImportError:
            raise ImportError("Settings not found '{}'".format(path))

        return s

    @staticmethod
    def get(s, key, default=None):
        """
        Get a settings value from key. Try to get key as it is, transforming to lower case and to upper case. If not
        found, a default value will be returned.

        :param s: Settings module or object.
        :param key: Settings key.
        :param default: Default value.
        :return: Settings value. Default value if key is not found.
        """
        return getattr(s, key, getattr(s, key.lower(), getattr(s, key.upper(), default)))

    @reset
    def build_from_django(self):
        from django.conf import settings as django_settings

        self.build_from_module(django_settings)

    @reset
    def build_from_module(self, module=None):
        if isinstance(module, str):
            module = self.import_settings(module)

        # Builder args
        self.run_at_startup = self.get(module, 'task_dispatcher_run_at_startup', [])

settings = Settings()



# -*- coding: utf-8 -*-
try:
    from django.core.management.base import BaseCommand
except ImportError:
    BaseCommand = object

from task_dispatcher.commands import TaskDispatcherCommand


class Command(BaseCommand):
    """
    Wrapper that makes a Django command from TaskDispatcherCommand.
    """
    help = TaskDispatcherCommand.description

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self.command = TaskDispatcherCommand(parse_args=False)

    def add_arguments(self, parser):
        self.command._add_arguments(parser=parser)

    def handle(self, *args, **options):
        self.command.run(**options)



# -*- coding: utf-8 -*-
import os
import sys

sys.path[0] = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))

from task_dispatcher.commands import TaskDispatcherCommand  # noqa

__all__ = ['main']


def main():
    return TaskDispatcherCommand().run()


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-

import os
import shutil
import sys

from pip.download import PipSession
from pip.req import parse_requirements as requirements
from setuptools import setup, Command

BASE_DIR = os.path.abspath(os.path.dirname(__file__))

if sys.version_info[0] == 2:
    from codecs import open


def parse_requirements(f):
    return [str(r.req) for r in requirements(f, session=PipSession())]


# Read requirements
_requirements_file = os.path.join(BASE_DIR, 'requirements.txt')
_tests_requirements_file = os.path.join(BASE_DIR, 'requirements-tests.txt')
_REQUIRES = parse_requirements(_requirements_file)
_TESTS_REQUIRES = parse_requirements(_tests_requirements_file)

# Read description
with open(os.path.join(BASE_DIR, 'README.rst'), encoding='utf-8') as f:
    _LONG_DESCRIPTION = f.read()

_CLASSIFIERS = (
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Natural Language :: English',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Software Development :: Object Brokering',
)
_KEYWORDS = ' '.join([
    'python',
    'stream',
    'data',
    'dispatcher',
    'flow',
    'database',
    'cache',
    'celery',
    'producer',
    'consumer',
    'scheduler',
])

setup(
    name='task-dispatcher',
    version='1.4.1',
    description='Library that provides a system to generate tasks producers and consumers with ease.',
    long_description=_LONG_DESCRIPTION,
    author='José Antonio Perdiguero López',
    author_email='perdy.hh@gmail.com',
    maintainer='José Antonio Perdiguero López',
    maintainer_email='perdy.hh@gmail.com',
    url='https://github.com/PeRDy/task-dispatcher',
    download_url='https://github.com/PeRDy/task-dispatcher',
    packages=[
        'task_dispatcher',
    ],
    include_package_data=True,
    install_requires=_REQUIRES,
    tests_require=_TESTS_REQUIRES,
    extras_require={
        'dev': [
                   'setuptools',
                   'pip',
                   'wheel',
                   'twine',
                   'bumpversion',
                   'pre-commit',
                   'sphinx',
                   'sphinx_rtd_theme'
               ] + _TESTS_REQUIRES,
    },
    license='GPLv3',
    zip_safe=False,
    keywords=_KEYWORDS,
    classifiers=_CLASSIFIERS,
    entry_points={
        'console_scripts': [
            'task_dispatcher = task_dispatcher.__main__:main',
        ],
    },
)
